from traitlets import Unicode
from IPython.display import display
import ipyvuetify as v 

# change vuetify theming 
v.theme.dark = True

# set the colors for the dark theme
v.theme.themes.dark.primary = "#B3842E"
v.theme.themes.dark.accent = "#a1458e"
v.theme.themes.dark.secondary = "#324a88"
v.theme.themes.dark.success = "#3F802A"
v.theme.themes.dark.info = "#79B1C9"
v.theme.themes.dark.warning = "#b8721d"
v.theme.themes.dark.error = "#A63228"

# fixed colors 
sepal_main = '#24221F'
sepal_darker = '#1a1a1a'

# set the background 
bg_color = '#121212' if v.theme.dark else '#fff'

# Fixed styles to avoid leaflet maps overlap sepal widgets
class Styles(v.VuetifyTemplate):

    template=Unicode("""
    <style>
        .leaflet-pane {
            z-index : 2 !important;
        }
        .leaflet-top, .leaflet-bottom {
            z-index : 2 !important;
        }
        main.v-content {
            padding-top: 0px !important;
        }
    </style>
    """).tag(sync=True)
    
styles=Styles()
display(styles)

COMPONENTS = {
    'PROGRESS_BAR':{
        'color':'indigo',
    }
}

# default styling of the aoi layer
AOI_STYLE = { 
    "stroke": True,
    "color": v.theme.themes.dark.success,
    "weight": 2,
    "opacity": 1,
    "fill": True,
    "fillColor": v.theme.themes.dark.success,
    "fillOpacity": 0.4,
}


_folder = {'color':'amber', 'icon':'mdi-folder-outline'}
_table = {'color':'green accent-4','icon':'mdi-border-all'}
_vector = {'color':'deep-purple', 'icon':'mdi-vector-polyline'}
_other = {'color':'light-blue', 'icon':'mdi-file-outline'}
_parent = {'color':'white', 'icon':'mdi-folder-upload-outline'}
_image = {'color':'deep-purple', 'icon':'mdi-image-outline'}

ICON_TYPES = {
    '': _folder,
    '.csv': _table,
    '.txt': _table,
    '.tif': _image,
    '.tiff': _image,
    '.shp': _vector,
    '.geojson': _vector,
    'DEFAULT': _other,
    'PARENT': _parent,
}



__author__ = """Pierrick Rambaud"""
__email__ = "pierrick.rambaud49@gmail.com"
__version__ = "2.0.2"

# direct access to colors 
from sepal_ui.frontend import styles
import ipyvuetify as v
from types import SimpleNamespace

theme = v.theme.themes.dark if v.theme.dark else v.theme.themes.light

color = SimpleNamespace(
    bg = styles.bg_color,
    primary = theme.primary,
    accent = theme.accent,
    secondary = theme.secondary,
    success = theme.success,
    info = theme.info,
    warning = theme.warning,
    error = theme.error
)



from functools import partial
from datetime import datetime

from ipywidgets import jslink
import ipyvuetify as v
from deprecated import deprecated
from traitlets import Unicode, observe, directional_link, List, Bool

from sepal_ui.sepalwidgets.sepalwidget import SepalWidget, TYPES

class Divider(v.Divider, SepalWidget):
    """
    A custom Divider with the ability to dynamically change color
    Whenever the type\_ trait is modified, the divider class will change accordingly
    
    Args:
        class\_ (str, optional): the initial color of the divider 
        
    Attributes:
        type_ (str): the current color of the divider
    """
    # Added type_ trait to specify default divider color
    type_ = Unicode('').tag(sync=True)

    def __init__(self, class_='', **kwargs):
        self.class_ = class_
        super().__init__(**kwargs)

    @observe('type_')
    def add_class_type(self, change):
        """
        Change the color of the divider according to the type\_
        It is binded to the type\_ traitlet but can also be called manually
        
        Args:
            change (dict): the only useful key is 'new' which is the new required color
            
        Return: 
            self
        """
         
        type_= change['new']
        classes = self.class_.split(' ')
        existing = list(set(classes) & set(TYPES))
        if existing:
            classes[classes.index(existing[0])] = type_
            self.class_ = " ".join(classes)
        else:
            self.class_ += f' {type_}'
            
        return self
            
            
class Alert(v.Alert, SepalWidget):
    """
    A custom Alert widget. It is used as the output of all processes in the framework. 
    In the voila interfaces, print statement will not be displayed. 
    Instead use the sw.Alert method to provide information to the user.
    It's hidden by default.
    
    Args:
        type\_ (str, optional): The color of the Alert
    """
    
    def __init__(self, type_=None, **kwargs):

        self.text = True
        self.type = type_ if (type_ in TYPES) else TYPES[0]
        self.class_="mt-5"
        
        super().__init__(**kwargs)
    
        self.hide()
        
    def update_progress(self, progress, msg='Progress', bar_length=30):
        """
        Update the Alert message with a progress bar. This function will stay until we manage to use tqdm in the widgets
        
        Args:
            progress (float): the progress status in float [0, 1]
            msg (str, optionnal): The message to use before the progress bar 
            bar_length (int, optionnal): the length of the progress bar in characters 
            
        Return:
            self
        """
        
        # define the characters to use in the progress bar
        plain_char = '█'
        empty_char = ' '  
        
        # cast the progress to float
        progress = float(progress)
        if not (0 <= progress <= 1):
            raise ValueError(f"progress should be in [0, 1], {progress} given")
        
        # set the length parameter 
        block = int(round(bar_length * progress))

        # construct the message content
        text = f'|{plain_char * block + empty_char * (bar_length - block)}|'
    
        # add the message to the output
        self.add_live_msg(v.Html(
            tag='span', 
            children=[
                v.Html(tag='span', children=[f'{msg}: '], class_='d-inline'),
                v.Html(tag='pre', class_='info--text d-inline', children=[text]),
                v.Html(tag='span', children=[f' {progress *100:.1f}%'], class_='d-inline')
            ]
        ))
   
        return self
        
    def add_msg(self, msg, type_='info'):
        """
        Add a message in the alert by replacing all the existing one. 
        The color can also be changed dynamically
        
        Args:
            msg (str): the message to display
            type\_ (str, optional): the color to use in the widget
            
        Return:
            self
        """
        self.show()
        self.type = type_ if (type_ in TYPES) else TYPES[0]
        self.children = [v.Html(tag='p', children=[msg])]
        
        return self
        
    def add_live_msg(self, msg, type_='info'):
        """
        Add a message in the alert by replacing all the existing one. 
        Also add the timestamp of the display.
        The color can also be changed dynamically
        
        Args:
            msg (str): the message to display
            type\_ (str, optional): the color to use in the widget
            
        Return:
            self
        """
        
        current_time = datetime.now().strftime("%Y/%m/%d, %H:%M:%S")

        self.show()
        self.type = type_ if (type_ in TYPES) else TYPES[0]
    
        self.children = [
            v.Html(tag='p', children=['[{}]'.format(current_time)]),
            v.Html(tag='p', children=[msg])
        ]
        
        return self

    def append_msg(self, msg, section=False, type_='info'):
        """ 
        Append a message in a new parragraph, with or without divider

        Args: 
            msg (str): the message to display
            section (bool, optional): add a Divider before the added message
            
        Return:
            self
        """
        
        self.show()
        self.type=type_
        
        if len(self.children):
            current_children = self.children[:]
            if section:
                # As the list is mutable, and the trait is only triggered when
                # the children is changed, so have to create a copy.
                divider = Divider(class_='my-4', style='opacity: 0.22')

                # link Alert type with divider type
                directional_link((self, 'type'), (divider, 'type_'))
                current_children.extend([
                    divider,
                    v.Html(tag='p', children=[msg])
                ])
                self.children = current_children

            else:
                current_children.append(
                    v.Html(tag='p', children=[msg])
                )
                self.children = current_children
        else:
            self.add_msg(msg)
            
        return self

    def remove_last_msg(self):
        """ 
        Remove the last msg printed in the Aler widget
        
        Return:
            self
        """
        
        if len(self.children) > 1:
            current_children = self.children[:]
            self.children = current_children[:-1]
        else:
            self.reset()
            
        return self

    def reset(self):
        """
        Empty the messages and hide it
        
        Return:
            self
        """
        
        self.children = ['']
        self.hide()
        
        return self
    
    @deprecated(reason="will be removed in version 2.1")
    def bind(self, widget, obj, attribute, msg=None, verbose=True, secret=False):
        """ 
        Bind the attribute to the widget and display it in the alert.
        The binded input need to have an active `v_model` trait.
        After the binding, whenever the `v_model` of the input is changed, the io attribute is changed accordingly.
        The value can also be displayed in the alert with a custom message with the following format = `[custom message] + [v_model]`
    
        Args:
            widget (v.XX): an ipyvuetify input element with an activated `v_model` trait
            obj (io): any io object
            attribute (str): the name of the attribute in io object
            msg (str, optionnal): the output message displayed before the variable
            verbose (bool, optional): wheter the variable should be displayed to the user 
            secret (bool, optional): either if the variable is secret or not. If true only "*" will be shown in the output
            
        Return:
            self
        """
        if not msg: msg = 'The selected variable is: '
        
        def _on_change(change, obj=obj, attribute=attribute, msg=msg):
            
            # if the key doesn't exist the getattr function will raise an AttributeError
            getattr(obj, attribute)
            
            # change the obj value
            setattr(obj, attribute, change['new'])
            
            # add the message if needed
            if secret:
                msg += '*' * len(str(change['new']))
            else:
                msg += str(change['new'])
                
            if verbose:
                self.add_msg(msg)
        
            return
        
        widget.observe(_on_change, 'v_model')
        
        return self
    
    
    def check_input(self, input_, msg=None):
        """
        Check if the inpupt value is initialized. 
        If not return false and display an error message else return True
        
        Args:
            input\_ (any): the input to check
            msg (str, optionnal): the message to display if the input is not set
            
        Return:
            (bool): check if the value is initialized
        """
        if not msg: msg = "The value has not been initialized"
        init = True 
        
        # check the collection type that are the only one supporting the len method
        try:
            if len(input_) == 0:
                init = False
        except:
            if input_ == None:
                init = False
                
        if not init:
            self.add_msg(msg, 'error')
        
        return init
    
class StateBar(v.SystemBar):
    
    """ Widget to display quick messages on simple inline status bar
        
    Attributes:
        msg (Unicode): the msg to be displayed 
        loading (Bool): State of bar, it will display a loading spin wheel if not loading.
    
    """
    
    msg = Unicode('').tag(sync=True)
    loading = Bool(False).tag(sync=True)
    
    def __init__(self,  **kwargs):
                        
        self.progress = v.ProgressCircular(
            indeterminate=self.loading,
            value=100,
            small=True,
            size=15,
            color='primary',
            class_='mr-2',
        )
        
        self.children = [self.progress, self.msg]
        
        super().__init__(**kwargs)
        
        jslink((self, 'loading'), (self.progress, 'indeterminate'))
    
    @observe('loading')
    def _change_loading(self, change):
        """ Change progress wheel state"""
        self.progress.indeterminate = self.loading
            
    @observe('msg')
    def _change_msg(self, change):
        """ Change state bar message"""
        self.children = [self.progress, self.msg]
        
    def add_msg(self, msg, loading=False):
        """ Change current status message"""
        self.msg = msg
        self.loading = loading
        
        return self



from pathlib import Path
import json

import ipyvuetify as v
from traitlets import link, Int, Any
from ipywidgets import jslink
import pandas as pd
import ee
import geopandas as gpd

from sepal_ui.frontend.styles import *
from sepal_ui.scripts import utils as su
from sepal_ui.scripts import gee
from sepal_ui.sepalwidgets.sepalwidget import SepalWidget
from sepal_ui.sepalwidgets.btn import Btn

class DatePicker(v.Layout, SepalWidget):
    """
    Custom input widget to provide a reusable DatePicker. It allows to choose date as a string in the following format YYYY-MM-DD
    
    Args:
        label (str, optional): the label of the datepicker field
        
    Attributes:
        menu (v.Menu): the menu widget to display the datepicker
        
    """
    def __init__(self, label="Date", **kwargs):
        
        date_picker = v.DatePicker(
            no_title = True, 
            v_model = None, 
            scrollable = True
        )

        date_text = v.TextField(
            v_model = None,
            label = label,
            hint = "YYYY-MM-DD format",
            persistent_hint = True, 
            prepend_icon = "event",
            readonly = True,
            v_on = 'menuData.on'
        )

        self.menu = v.Menu(
            min_width="290px",
            transition = "scale-transition",
            offset_y = True, 
            value = False,
            close_on_content_click = False,
            children = [date_picker],
            v_slots = [{
                'name': 'activator',
                'variable': 'menuData',
                'children': date_text,
            }]
        )

        super().__init__(
            v_model = None,
            row = True,
            class_ = 'pa-5',
            align_center = True,
            children = [v.Flex(xs10=True, children=[self.menu])],
            **kwargs
        )

        jslink((date_picker, 'v_model'), (date_text, 'v_model'))
        jslink((date_picker, 'v_model'), (self, 'v_model'))
        
        #close the datepicker on click
        #date_text.observe(lambda _: setattr(self.menu, 'value', False), 'v_model')
        
class FileInput(v.Flex, SepalWidget):
    """
    Custom input field to select a file in the sepal folders. 
    
    Args:
        extentions ([str]): the list of the allowed extentions. the FileInput will only display these extention and folders
        folder (str | pathlib.Path): the starting folder of the file input
        label (str): the label of the input
        v_model (str, optional): the default value
        
    Attributes:
        extentions ([str]): the extention list
        folder (str | pathlib.Path): the current folder
        selected_file (v.TextField): the textfield where the file pathname is stored
        loading (v.ProgressLinear): loading top bar of the menu component
        file_list (v.List): the list of files and folder that are available in the current folder
        file_menu (v.Menu): the menu that hide and show the file_list
        reload (v.Btn): reload btn to reload the file list on the current folder 
    """

    file = Any('')
    
    def __init__(self, 
                 extentions = [], 
                 folder=Path('~').expanduser(), 
                 label='search file', 
                 v_model = None, 
                 **kwargs):
        
        if type(folder) == str:
            folder = Path(folder)
            
        self.extentions = extentions
        self.folder = folder
        self.v_model = v_model
        
        self.selected_file = v.TextField(
            readonly = True,
            label    = 'Selected file', 
            class_   = 'ml-5 mt-5',
            v_model  = self.file
        )

        self.loading = v.ProgressLinear(
            indeterminate    = False, 
            background_color = 'grey darken-3',
            color            = COMPONENTS['PROGRESS_BAR']['color']
            )
        
        self.file_list = v.List(
            dense      = True, 
            color      = 'grey darken-3',
            flat       = True,
            max_height = '300px',
            style_     = 'overflow: auto; border-radius: 0 0 0 0;',
            children   = [ 
                v.ListItemGroup(
                    children = self._get_items(),
                    v_model  = ''
                )
            ]
        )

        self.file_menu = v.Menu(
            min_width              = 300,
            children               = [self.loading, self.file_list], 
            close_on_content_click = False,
            v_slots                = [{
                'name': 'activator',
                'variable': 'x',
                'children': Btn(icon='mdi-file-search', v_model=False, v_on='x.on', text=label)
        }])
        
        self.reload = v.Btn(
            icon = True,
            color = 'primary',
            children = [v.Icon(children=['mdi-cached'])]
        )
        
        super().__init__(
            row          = True,
            class_       = 'd-flex align-center mb-2',
            align_center = True,
            children     = [
                self.reload,
                self.file_menu,
                self.selected_file,
            ],
            **kwargs
        )
        
        link((self.selected_file, 'v_model'), (self, 'file'))
        link((self.selected_file, 'v_model'), (self, 'v_model'))

        self.file_list.children[0].observe(self._on_file_select, 'v_model')
        self.reload.on_event('click', self._on_reload)
        
    def reset(self):
        """
        Clear the File selection and move to the root folder if something was selected
        
        Return:
            self
        """
        
        root = Path('~').expanduser()
        
        if self.v_model != '':
            
            # move to root
            self._on_file_select({'new': root})
        
            # remove v_model
            self.v_model = ''
        
        return self
    
    def select_file(self, path):
        """
        Manually select a file from it's path. no verification on the extension is performed
        
        Params:
            path (str|pathlib.Path): the path to the file
        
        Return:
            self
        """
        
        # cast to Path
        path = Path(path)
        
        # test file existence 
        if not path.is_file():
            raise Exception(f'{path} is not a file')
        
        # set the menu to the folder of the file
        self._on_file_select({'new': path.parent})
        
        # select the appropriate file 
        self._on_file_select({'new': path})
        
    def _on_file_select(self, change):
        """Dispatch the behavior between file selection and folder change"""
        
        if not change['new']:
            return self
        
        new_value = Path(change['new'])
        
        if new_value.is_dir():
            self.folder = new_value
            self._change_folder()
                
        elif new_value.is_file():
            self.file = str(new_value)
            
        return self
                
    def _change_folder(self):
        """Change the target folder"""
        #reset files
        self.file_list.children[0].children = self._get_items()
        
        return
    
    def _get_items(self):
        """Return the list of items inside the folder"""

        self.loading.indeterminate = True
        
        folder = self.folder

        list_dir = [el for el in folder.glob('*/') if not el.name.startswith('.')]

        if self.extentions:
            list_dir = [el for el in list_dir if el.is_dir() or el.suffix in self.extentions]

        folder_list = []
        file_list = []

        for el in list_dir:
            
            if el.suffix in ICON_TYPES.keys():
                icon = ICON_TYPES[el.suffix]['icon']
                color = ICON_TYPES[el.suffix]['color']
            else:
                icon = ICON_TYPES['DEFAULT']['icon']
                color = ICON_TYPES['DEFAULT']['color']
            
            children = [
                v.ListItemAction(children=[v.Icon(color= color,children=[icon])]),
                v.ListItemContent(children=[v.ListItemTitle(children=[el.stem + el.suffix])]),
            ] 

            if el.is_dir():
                folder_list.append(v.ListItem(value=str(el), children=children))
            else:
                file_size = su.get_file_size(el)
                children.append(v.ListItemActionText(children=[file_size]))
                file_list.append(v.ListItem(value=str(el), children=children))

        folder_list = sorted(folder_list, key=lambda x: x.value)
        file_list = sorted(file_list, key=lambda x: x.value)

        parent_item = v.ListItem(
            value=str(folder.parent), 
            children=[
                v.ListItemAction(children=[v.Icon(color=ICON_TYPES['PARENT']['color'], children=[ICON_TYPES['PARENT']['icon']])]),
                v.ListItemContent(children=[v.ListItemTitle(children=[f'..{folder.parent}'])]),
            ]
        )

        folder_list.extend(file_list)
        folder_list.insert(0,parent_item)

        self.loading.indeterminate = False
        
        return folder_list
    
    def _on_reload(self, widget, event, data):
        
        # force the update of the current folder
        self._change_folder()
        
        return

class LoadTableField(v.Col, SepalWidget):
    """
    A custom input widget to load points data. The user will provide a csv or txt file containing labeled dataset. 
    The relevant columns (lat, long and id) can then be identified in the updated select. Once everything is set, the widget will populate itself with a json dict.
    {pathname, id_column, lat_column,lng_column}
    
    Attributes:
        fileInput (sw.FileInput): the file input to select the .csv or .txt file
        v_model (Traitlet): the json saved v_model shaped as {'pathname': xx, 'id_column': xx, 'lat_column': xx, 'lng_column': xx} The values can be accessed separately with the appropriate getter methods
        IdSelect (v.Select): input to select the id column
        LngSelect (v.Select): input to select the lng column
        LatSelect (v.Select): input to select the lat column
    """
    
    default_v_model = {
        'pathname'  : None, 
        'id_column' : None, 
        'lat_column': None, 
        'lng_column': None
    }
    
    def __init__(self, label="Table file"):
        
        self.fileInput = FileInput(['.csv', '.txt'], label=label)
        
        self.IdSelect = v.Select(
            _metadata = {'name': 'id_column'}, 
            items     = [], 
            label     = 'Id', 
            v_model   = None
        ) 
        self.LngSelect = v.Select(
            _metadata = {'name': 'lng_column'}, 
            items     = [], 
            label     = 'Longitude', 
            v_model   = None
        )
        self.LatSelect = v.Select(
            _metadata = {'name': 'lat_column'}, 
            items     = [], 
            label     = 'Latitude', 
            v_model   = None
        )
        
        super().__init__(
            v_model = self.default_v_model,
            children = [
                self.fileInput,
                self.IdSelect,
                self.LngSelect,
                self.LatSelect
            ]
        )
        
        # link the dropdowns
        jslink((self.IdSelect, 'items'),(self.LngSelect, 'items'))
        jslink((self.IdSelect, 'items'),(self.LatSelect, 'items'))
        
        # link the widget with v_model 
        self.fileInput.observe(self._on_file_input_change, 'v_model')
        self.IdSelect.observe(self._on_select_change, 'v_model')
        self.LngSelect.observe(self._on_select_change, 'v_model')
        self.LatSelect.observe(self._on_select_change, 'v_model')
        
    def reset(self):
        """
        Clear the values and return to the empty default json
        
        Return:
            self
        """
        
        # clear the fileInput
        self.fileInput.reset()
        
    def _on_file_input_change(self, change):
        """Update the select content when the fileinput v_model is changing"""
        
        # clear the selects
        self._clear_select()
        
        # set the path
        path = change['new']
        self.v_model['pathname'] = path
        
        # exit if none
        if not path:
            return self
        
        df = pd.read_csv(path, sep=None, engine='python')
        
        if len(df.columns) < 3: 
            self._clear_select()
            return 
        
        # set the items 
        self.IdSelect.items = df.columns.tolist()
        
        # pre load values that sounds like what we are looking for 
        # it will only keep the first occurence of each one 
        for name in reversed(df.columns.tolist()):
            lname = name.lower() 
            if 'id' in lname:
                self.IdSelect.v_model = name
            elif any(ext in lname for ext in ['lng', 'long', 'longitude', 'x_coord', 'xcoord', 'lon']):
                self.LngSelect.v_model = name
            elif any(ext in lname for ext in ['lat', 'latitude', 'y_coord', 'ycoord']):
                self.LatSelect.v_model = name
        
        return self
                
    def _clear_select(self):
        """clear the selects components"""
        
        self.IdSelect.items = [] # all the others are listening to this one 
        self.IdSelect.v_model = self.LngSelect.v_model = self.LatSelect.v_model = None
        
        return self
    
    def _on_select_change(self, change):
        """change the v_model value when a select is changed"""
        
        name = change['owner']._metadata['name']
        self.v_model[name] = change['new']
        
        return self

class AssetSelect(v.Combobox, SepalWidget):
    """
    Custom widget input to select an asset inside the asset folder of the user
    
    Args:
        label (str): the label of the input
        folder (str): the folder of the user assets
        default_asset (str): the id of a default asset
        
    Attributes:
        folder (str): the folder of the user assets
    """
    

    @su.need_ee
    def __init__(self, label = 'Select an asset', folder = None, default_asset = None, **kwargs):
        
        # initialize earth engine
        su.init_ee()
        
        # if folder is not set use the root one 
        self.folder = folder if folder else ee.data.getAssetRoots()[0]['id']
        
        self.label = label
        self.v_model = default_asset
        
        self.clearable = True
        self.dense = True
        self.persistent_hint = True
        self.prepend_icon = 'mdi-cached'
        
        self.class_ = 'my-5'
        self.placeholder = 'users/someCustomUser/customAsset'
        self.hint = 'select an asset in the list or write a custom asset name. Be careful that you need to have access to this asset to use it'
        
        self._get_items(None, None, None)
        
        super().__init__(**kwargs) 
        
        self.on_event('click:prepend', self._get_items)

    def _get_items(self, widget, event, data):
        
        # get the list of user asset
        assets = gee.get_assets(self.folder)
        
        tables = sorted([e['id'] for e in assets if e['type'] == 'TABLE'])
        images = sorted([e['id'] for e in assets if e['type'] == 'IMAGE'])
        
        self.items = [
            {'divider':True}, {'header':'Tables'},
            *tables,
            {'divider':True}, {'header':'Rasters'},
            *images
        ]
        
        return self
        
class PasswordField(v.TextField, SepalWidget):
    """
    Custom widget to input passwords in text area and 
    toggle its visibility.

    Args:
        label (str, optional): Header displayed in text area. Defaults to Password.
    """
    def __init__(self, **kwargs):
        
        # default behavior 
        self.label="Password"
        self.class_='mr-2'
        self.v_model=''
        self.type='password'
        self.append_icon='mdi-eye-off'
        
        # init the widget with the remaining kwargs
        super().__init__(**kwargs)
        
        # bind the js behavior
        self.on_event('click:append' ,self._toggle_pwd)
    

    def _toggle_pwd(self, widget, event, data):
        """Toggle password visibility when append button is clicked"""
        
        if self.type=='text':
            self.type='password'
            self.append_icon = 'mdi-eye-off'
        else:
            self.type = 'text'
            self.append_icon = 'mdi-eye'
            
class NumberField(v.TextField, SepalWidget):
    """
    Custom widget to input numbers in text area and add/substract with single increment.

    Args:
        max_ (int, optional): Maximum selectable number. Defaults to 10.
        min_ (int, optional): Minimum selectable number. Defaults to 0.

    """
    max_ = Int(10).tag(sync=True)
    min_ = Int(0).tag(sync=True)
    
    def __init__(self, **kwargs):
        
        self.type='number'
        self.append_outer_icon='mdi-plus'
        self.prepend_icon='mdi-minus'
        self.v_model=0
        self.readonly=True
        
        super().__init__(**kwargs)
        
        self.on_event('click:append-outer', self.increment)
        self.on_event('click:prepend', self.decrement)
    
    def increment(self, widget, event, data):
        """Adds 1 to the current v_model number"""
        if self.v_model < self.max_: self.v_model+=1
        
    def decrement(self, widget, event, data):
        """Substracts 1 to the current v_model number"""
        if self.v_model > self.min_: self.v_model-=1

class VectorField(v.Col, SepalWidget):
    """
    A custom input widget to load vector data. The user will provide a vector file compatible with fiona.
    The user can then select a specific shape by setting column and value fields.
    
    Args:
        label (str): the label of the file input field, default to 'vector file'.
    
    Attributes:
        original_gdf (geopandas.gdf): The originally selected dataframe
        gdf (geopandas.gdf): The selected dataframe.
        v_model (Traitlet): The json saved v_model shaped as {'pathname': xx, 'column': xx, 'value': xx} The values can be accessed separately with the appropriate getter methods
        w_file (sw.FileInput): The file selector widget
        w_column (v.Select): The Select widget to select the column
        w_value (v.Select): The Select widget to select the value in the selected column 
    """
    
    default_v_model = {
        'pathname'  : None, 
        'column' : None, 
        'value': None, 
    }
    
    column_base_items = [
        {'text': 'Use all features', 'value': 'ALL'},
        {'divider': True}
    ]
        
    def __init__(self, label='vector_file', **kwargs):
        
        # save the df for column naming (not using a gdf as geometry are useless)
        self.df = None
        
        # set the 3 wigets
        self.w_file = FileInput(['.shp', '.geojson', '.gpkg', '.kml'], label=label)
        self.w_column = v.Select(
            _metadata = {'name': 'column'}, 
            items     = self.column_base_items, 
            label     = 'Column', 
            v_model   = 'ALL'
        )
        self.w_value = v.Select(
            _metadata = {'name': 'value'}, 
            items     = [], 
            label     = 'Value', 
            v_model   = None
        )
        su.hide_component(self.w_value)
        
        
        # create the Col Field
        self.children = [self.w_file, self.w_column, self.w_value]
        self.v_model = self.default_v_model
        
        super().__init__(**kwargs)
        
        # events
        self.w_file.observe(self._update_file, 'v_model')
        self.w_column.observe(self._update_column, 'v_model')
        self.w_value.observe(self._update_value, 'v_model')
        
    def reset(self):
        """
        Return the field to its initial state
        
        Return:
            self
        """
        
        self.w_file.reset()
        
        return self
        
    def _update_file(self, change):
        """update the file name, the v_model and reset the other widgets"""
        
        # reset the widgets
        self.w_column.items = self.w_value.items = []
        self.w_column.v_model = self.w_value.v_model = None
        self.df = None
        
        # set the pathname value 
        self.v_model["pathname"] = change['new']
        
        # exit if nothing 
        if not change['new']: return self
        
        # read the file 
        self.df = gpd.read_file(change['new'], ignore_geometry=True)        

        # update the columns
        self.w_column.items = self.column_base_items + sorted(set(self.df.columns.to_list()))
        
        self.w_column.v_model = 'ALL'
        
        return self
    
    def _update_column(self, change):
        """Update the column name and empty the value list"""
        
        # reset value widget
        self.w_value.items = []
        self.w_value.v_model = None
        
        # set the value 
        self.v_model['column'] = change['new']
        
        # hide value if "ALL" or none
        if change['new'] in ['ALL', None]:
            su.hide_component(self.w_value)
            return self
        
        # read the colmun 
        self.w_value.items = sorted(set(self.df[change['new']].to_list()))
        su.show_component(self.w_value)
        
        return self
    
    def _update_value(self, change):
        """Update the value name and reduce the gdf"""
        
        # set the value 
        self.v_model['value'] = change['new']
        
        return self    
    



import time

import ee

from sepal_ui.message import ms
from sepal_ui.scripts import utils as su

@su.need_ee
def wait_for_completion(task_descripsion, widget_alert=None):
    """
    Wait until the selected process is finished. Display some output information

    Args:
        task_descripsion (str): name of the running task
        widget_alert (v.Alert, optional): alert to display the output messages
        
    Return:
        (str): the final state of the task
    """
    state = 'UNSUBMITTED'
    while state != 'COMPLETED':
        
        # print in a widget
        if widget_alert: 
            widget_alert.add_live_msg(ms.status.format(state))
            
        # wait 5 seconds
        time.sleep(5)
                    
        # search for the task in task_list
        current_task = is_task(task_descripsion)
        state = current_task.state
        
        if state == 'FAILED':
            raise Exception(ms.status.format(state))
        
    # print in a widget
    if widget_alert: 
        widget_alert.add_live_msg(ms.status.format(state), 'success')
        
    return state

@su.need_ee
def is_task(task_descripsion):
    """
    Search for the described task in the user Task list return None if nothing is found
    
    Args: 
        task_descripsion (str): the task description
    
    Return:
        (ee.Task) : return the found task else None
    """
    
    current_task = None
    for task in ee.batch.Task.list():
        if task.config['description'] == task_descripsion:
            current_task = task
            break
            
    return current_task

@su.need_ee
def is_running(task_descripsion):
    """
    Search for the described task in the user Task list return None if nothing is currently running
    
    Args: 
        task_descripsion (str): the task description
    
    Return:
        (ee.Task) : return the found task else None
    """
    
    current_task = is_task(task_descripsion)
    if current_task:
        if current_task.state != 'RUNNING':
            current_task = None
            
    return current_task
    
@su.need_ee
def get_assets(folder=None, asset_list = []):
    """
    Get all the assets from the parameter folder. every nested asset will be displayed.
    
    Args:
        folder (str): the initial GEE folder 
        asset_list ([assets]| optional): extra element that you would like to add to the asset list
        
    Return:
        ([asset]): the asset list. each asset is a dict with 3 keys: 'type', 'name' and 'id'
    """
    # set the folder
    folder = folder if folder else ee.data.getAssetRoots()[0]['id']
    
    # loop in the assets
    for asset in ee.data.listAssets({'parent': folder})['assets']:
        if asset['type'] == 'FOLDER':
            asset_list = get_assets(asset['name'], asset_list)
        else:
            asset_list += [asset]
    
    return asset_list

@su.need_ee
def is_asset(asset_name, folder=None):
    """
    Check if the asset already exist in the user asset folder
    
    Args: 
        asset_descripsion (str) : the descripsion of the asset
        folder (str): the folder of the glad assets
        
    Return:
        (bool): true if already in folder
    """
    
    # get the folder 
    folder = folder or ee.data.getAssetRoots()[0]['id']
    
    # get all the assets
    asset_list = get_assets(folder)
    
    # search for asset existance
    exist = False
    for asset in asset_list:
        if asset_name == asset['name']:
            exist = True
            break
            
    return exist



from pathlib import Path
from traitlets import Int
from datetime import datetime as dt

import ipyvuetify as v
import pandas as pd
import geopandas as gpd
from shapely import geometry as sg

import sepal_ui.sepalwidgets as sw
from sepal_ui.scripts import utils as su
from sepal_ui.aoi.aoi_model import AoiModel
from sepal_ui.message import ms
from sepal_ui import color as sc

CUSTOM = AoiModel.CUSTOM
ADMIN = AoiModel.ADMIN
ALL = 'All'
select_methods = AoiModel.METHODS

class Select(v.Select, sw.SepalWidget):
    """A classic Vuetify Select widget inheriting from sepalwidgets"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
class TextField(v.TextField, sw.SepalWidget):
    """ A classic Vuetify TextField widget inheriting from sepalwidgets"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

class MethodSelect(Select):
    f"""
    A method selector. It will list the availabel methods for this very AoiView. 
    'ALL' will select all the available methods (default)
    'ADMIN' only the admin one, 'CUSTOM' only the custom one. 
    'XXX' will add the selected method to the list when '-XXX' will discard it. 
    You cannot mix adding and removing behaviours.
    
    Params:
        methods (str|[str]): a list of methods from the available list ({' '.join(select_methods.keys())})
    """
    
    def __init__(self, methods='ALL'):
        
        # create the method list
        if methods=='ALL':
            self.methods = select_methods
        elif methods == 'ADMIN':
            self.methods = {k: v for k, v in select_methods.items() if v['type'] == ADMIN}
        elif methods == 'CUSTOM':
            self.methods = {k: v for k, v in select_methods.items() if v['type'] == CUSTOM}
        elif type(methods) == list:
            
            if any(m[0] == '-' for m in methods) != all(m[0] == '-' for m in methods):
                raise Exception("You mixed adding and removing, punk")
            
            if methods[0][0] == '-':
                
                to_remove = [method[1:] for method in methods]
                
                # Rewrite the methods instead of mutate the class methods
                self.methods = {k:v for k,v in select_methods.items() if k not in to_remove} 
                
            else:
                self.methods = {k: select_methods[k] for k in methods}
        else:
            raise Exception("I don't get what you meant")
            
        # build the item list with header 
        prev_type = None
        items = []
        for k, m in self.methods.items():
            current_type = m['type']
            
            if prev_type != current_type:
                items.append({'header': current_type})
            prev_type = current_type
            
            items.append({'text':m['name'], 'value': k})
            
        # create the input 
        super().__init__(label=ms.aoi_sel.method, items=items, v_model='', dense=True)
        
class AdminField(v.Select, sw.SepalWidget):
    """
    An admin level selector. It is binded to ee (GAUL 2015) or not (GADM 2021). allows to select administrative codes taking into account the administrative parent code and displaying humanly readable administrative names.
    
    Args:
        level (int): The administrative level of the field
        parent (AdminField): the adminField that deal with the parent admin level of the current selector. used to narrow down the possible options
        ee (bool, optional): wether to use ee or not (default to True)
        
    Attributes:
        gee (bool): the earthengine status
        level (int): the admin level of the current field
        parent (AdminField): the field parent object
    """
    
    # the file location of the database
    FILE = AoiModel.FILE
    CODE = AoiModel.CODE
    NAME = AoiModel.NAME
    
    def __init__(self, level, parent=None, gee=True, **kwargs):
        
        self.toto = "toto"
        
        # save ee state
        self.ee = gee
        
        # get the level info 
        self.level = level
        self.parent = parent
        
        # init an empty widget
        self.v_model = None
        self.items = []
        self.clearable = True
        self.label = ms.aoi_sel.adm[level]
        super().__init__(**kwargs)
        
        # add js behaviour
        if self.parent:
            self.parent.observe(self._update, 'v_model')
            
    def show(self):
        """
        when an admin field is shown, show its parent as well
        
        Return:
            self
        """
        
        super().show()
        
        if self.parent: self.parent.show()
        
        return self
            
    def get_items(self, filter_=None):
        """
        update the item list based on the given filter
        
        Params:
            filter_ (str): The code of the parent v_model to filter the current results
            
        Return:
            self
        """
        
        # extract the level list
        df = pd.read_csv(self.FILE[self.ee]) \
            .drop_duplicates(subset=self.CODE[self.ee].format(self.level)) \
            .sort_values(self.NAME[self.ee].format(self.level))
        
        # filter it 
        if filter_: df = df[df[self.CODE[self.ee].format(self.level-1)] == filter_]
        
        # formatted as a item list for a select component
        self.items = [
            {
                'text': su.normalize_str(r[self.NAME[self.ee].format(self.level)], folder=False), 
                'value': r[self.CODE[self.ee].format(self.level)]
            } for _, r in df.iterrows()
        ] 
        
        return self
        
    def _update(self, change):
        """update the item list of the admin select"""
        
        # reset v_model
        self.v_model = None
        
        # update the items list
        if change['new']:
            self.get_items(change['new'])
            
        return self

class AoiView(v.Card):
    """
    Versatile card object to deal with the aoi selection. multiple selection method are available (see the MethodSelector object) and the widget can be fully customizable. Can also be bound to ee (ee==True) or not (ee==False)
    
    Args:
        methods (str, optional): the methods to use in the widget, default to 'ALL',
        map_ (SepalMap, optional): link the aoi_view to a custom SepalMap to display the output, default to None
        gee (bool, optional): wether to bind to ee or not
        vector (str|pathlib.Path, optional): the path to the default vector object
        admin (int, optional): the administrative code of the default selection. Need to be GADM if ee==False and GAUL 2015 if ee==True.
        asset (str, optional): the default asset. Can only work if ee==True
        
    Attributes:
        gee (bool): the earthengine binding status
        model (AoiModel): the aoiModel to save the aoi selected as gdf or ee.FeatureCollection
        map (SepalMap): the map if filled in the arguments
        w_method (MethodSelect): the the method selection widget
        w_admin_0, w_admin_1, w_admin2, w_vector, w_points, w_draw, w_asset (widgets): the selection widgets
        components (dict): the widgets stored in a dict using the method names as keys
        alert (sw.Alert): an alert component to display outputs
        btn (sw.Btn): a btn component
        
    """
    
    updated = Int(0).tag(sync=True)
    
    def __init__(self, methods='ALL', map_=None, gee=True, folder=None, **kwargs):
        
        # set ee dependencie
        self.ee = gee
        if gee: 
            su.init_ee()
            self.folder = folder
            
        # get the model   
        self.model = AoiModel(sw.Alert(), gee=gee, folder=folder, **kwargs)
        
        # get the map if filled 
        self.map_=map_
        
        # create the method widget 
        self.w_method = MethodSelect(methods)
        
        # add the 6 methods blocks
        self.w_admin_0 = AdminField(0, gee=gee).get_items().hide()
        self.w_admin_1 = AdminField(1, self.w_admin_0, gee=gee).hide()
        self.w_admin_2 = AdminField(2, self.w_admin_1, gee=gee).hide()
        self.w_vector = sw.VectorField(label=ms.aoi_sel.vector).hide()
        self.w_points = sw.LoadTableField(label=ms.aoi_sel.points).hide()
        if self.map_: self.w_draw = TextField(label=ms.aoi_sel.aoi_name).hide()
        if self.ee: self.w_asset = sw.AssetSelect(label=ms.aoi_sel.asset, folder=self.folder).hide()
        
        # group them together with the same key as the select_method object
        self.components = {
            'ADMIN0': self.w_admin_0,
            'ADMIN1': self.w_admin_1,
            'ADMIN2': self.w_admin_2,
            'SHAPE': self.w_vector,  
            'POINTS': self.w_points  
        }
        if self.map_: self.components['DRAW'] = self.w_draw
        if self.ee: self.components['ASSET'] = self.w_asset
        
        # use the same alert as in the model
        self.alert = self.model.alert
        
        # bind the widgets to the model
        self.model \
            .bind(self.w_admin_0, 'admin') \
            .bind(self.w_admin_1, 'admin') \
            .bind(self.w_admin_2, 'admin') \
            .bind(self.w_vector, 'vector_json') \
            .bind(self.w_points, 'point_json') \
            .bind(self.w_method, 'method')
        if self.map_: self.model.bind(self.w_draw, 'name')
        if self.ee: self.model.bind(self.w_asset, 'asset_name')
            
        # add a validation btn
        self.btn = sw.Btn(ms.aoi_sel.btn)
        
        # create the widget
        self.children = [self.w_method] + [*self.components.values()] + [self.btn, self.alert]
        
        super().__init__(**kwargs)
        
        # js events
        self.w_method.observe(self._activate, 'v_model') # activate the appropriate widgets
        self.btn.on_event('click', self._update_aoi) # load the informations
        if self.map_: self.map_.dc.on_draw(self._handle_draw) # handle map drawing
            
    @su.loading_button(debug=False)
    def _update_aoi(self, widget, event, data):
        """load the object in the model & update the map (if possible)"""
        
        # update the model 
        self.model.set_object()
        
        # update the map
        if self.map_:
            [self.map_.remove_layer(l) for l in self.map_.layers if l.name == 'aoi']
            self.map_.zoom_bounds(self.model.total_bounds())
            
            if self.ee:
                self.map_.addLayer(self.model.feature_collection, {'color': sc.success}, 'aoi')
            else:
                self.map_.add_layer(self.model.get_ipygeojson())
                
            self.map_.hide_dc()
        
        # tell the rest of the apps that the aoi have been updated 
        self.updated += 1
        
        return self
    
    def reset(self):
        """clear the aoi_model from input and remove the layer from the map (if existing)"""
        
        # clear the map
        if self.map_:
            [self.map_.remove_layer(l) for l in self.map_.layers if l.name == 'aoi']
            #self.map_.center = [0, 0]
            #self.map_.zoom = 3
            
        # clear the model
        self.model.clear_attributes()
        
        return self
            
        
    
    def _activate(self, change):
        """activate the adapted widgets"""
        
        # clear and hide the alert 
        self.alert.reset()
        
        # deactivate or activate the dc
        if self.map_: self.map_.show_dc() if change['new'] == 'DRAW' else self.map_.hide_dc()
    
        # clear the inputs
        [w.reset() for w in self.components.values()]
        
        # activate the widget
        [w.show() if change['new'] == k else w.hide() for k, w in self.components.items()]
        
        return self
    
    def _handle_draw(self, target, action, geo_json):
        """handle the draw on map event"""
        
        # update the automatic name
        if not self.w_draw.v_model:
            self.w_draw.v_model = f'Manual_aoi_{dt.now().strftime("%Y-%m-%d_%H-%M-%S")}'
            
        # Init the json if it's not 
        if self.model.geo_json == None:
            self.model.geo_json = {'type': 'FeatureCollection', 'features': []}
        
        # polygonize circles 
        if 'radius' in geo_json['properties']['style']:
            geo_json = self.polygonize(geo_json)
        
        if action == 'created': # no edit as you don't know which one to change
            self.model.geo_json['features'].append(geo_json)
        elif action == 'deleted':
            self.model.geo_json['features'].remove(geo_json)
            
        return self
    
    @staticmethod
    def polygonize(geo_json):
        """
        Transform a ipyleaflet circle (a point with a radius) into a GeoJson multipolygon
        
        Params:
            geo_json (json): the circle geojson
            
        Return:
            (json): the polygonised circle
        """
        
        # get the input
        radius = geo_json['properties']['style']['radius']
        coordinates = geo_json['geometry']['coordinates']
        
        # create shapely point 
        circle = gpd.GeoSeries([sg.Point(coordinates)], crs=4326).to_crs(3857).buffer(radius).to_crs(4326)
        
        # insert it in the geo_json 
        json = geo_json
        json['geometry'] = circle[0].__geo_interface__
        
        return json


import os
from pathlib import Path
from urllib.parse import urlparse
import string 
import random
import math
import re
import warnings
from unidecode import unidecode
from functools import wraps

import ee
from cryptography.fernet import Fernet

import sepal_ui

def hide_component(widget):
    """
    hide a vuetify based component
    
    Args:
        widget (v.VuetifyWidget): the widget to hide
    """
    
    if isinstance(widget, sepal_ui.sepalwidgets.sepalwidget.SepalWidget):
        widget.hide()
    elif not 'd-none' in str(widget.class_):
        widget.class_ = str(widget.class_).strip() + ' d-none'
        
    return

def show_component(widget):
    """
    show a vuetify based component
    
    Args:
        widget (v.VuetifyWidget): the widget to hide
    """
    
    if isinstance(widget, sepal_ui.sepalwidgets.sepalwidget.SepalWidget):
        widget.show()
    elif 'd-none' in str(widget.class_):
        widget.class_ = widget.class_.replace('d-none', '')
        
    return
    
def create_download_link(pathname):
    """
    Create a clickable link to download the pathname target
    
    Args:
        pathname (str | pathlib.Path): the pathname th download
        
    Return:
        (str): the download link
    """
    
    if type(pathname) == str:
        pathname = Path(pathname)
        
    result_path = Path(pathname).expanduser()
    home_path = Path('~').expanduser()
    
    # will be available with python 3.9
    #download_path = result_path.relative_to(home_path) if result_path.is_relative_to(home_path) else result_path
    download_path = os.path.relpath(result_path,home_path)
    
    link = f'/api/files/download?path=/{download_path}'
    
    return link

def is_absolute(url):
    """
    Check if the given URL is an absolute or relative path
    
    Args:
        url (str): the URL to test
        
    Return:
        (bool): True if absolute else False
    """
    return bool(urlparse(str(url)).netloc)

def random_string(string_length=3):
    """
    Generates a random string of fixed length. 
    
    Args:
        string_length (int, optional): Fixed length. Defaults to 3.
    
    Return:
        (str): A random string
    """
    
    # random.seed(1001)
    letters = string.ascii_lowercase
    
    return ''.join(random.choice(letters) for i in range(string_length))

def get_file_size(filename):
    """
    Get the file size as string of 2 digit in the adapted scale (B, KB, MB....)
    
    Args:
        filename (str | pathlib.Path): the path to the file to mesure
        
    Return:
        (str): the file size in a readable humanly readable
    """
    
    file_size = Path(filename).stat().st_size
    
    if file_size == 0:
        return "0B"
    
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    
    i = int(math.floor(math.log(file_size, 1024)))
    s = file_size / (1024 ** i)
        
    return '{:.1f} {}'.format(s, size_name[i])

def init_ee():
    """
    Initialize earth engine according to the environment. 
    It will use the creddential file if the EE_PRIVATE_KEY env variable exist. 
    Otherwise it use the simple Initialize command (asking the user to register if necessary)
    """
    
    # only do the initialization if the credential are missing
    if not ee.data._credentials:
        
        # if the decrypt key is available use the decript key 
        if 'EE_DECRYPT_KEY' in os.environ:
        
            # read the key as byte 
            key = os.environ['EE_DECRYPT_KEY'].encode()
            
            # create the fernet object 
            fernet = Fernet(key)
            
            # decrypt the key
            json_encrypted = Path(__file__).parent/'encrypted_key.json'
            with json_encrypted.open('rb') as f:
                json_decripted = fernet.decrypt(f.read()).decode()
                
            # write it to a file
            with open('ee_private_key.json', 'w') as f:
                f.write(json_decripted)
                
            # connection to the service account
            service_account = 'test-sepal-ui@sepal-ui.iam.gserviceaccount.com'
            credentials = ee.ServiceAccountCredentials(service_account, 'ee_private_key.json')
            ee.Initialize(credentials)
        
        # if in local env use the local user credential
        else:
            ee.Initialize()
        
    return

def catch_errors(alert, debug=False):
    """
    Decorator to execute try/except sentence
    and catch errors in the alert message.
    If debug is True then the error is raised anyway
    
    Params:
        alert (sw.Alert): Alert to display errors
        debug (bool): Wether to raise the error or not, default to false
    """
    def decorator_alert_error(func):
        @wraps(func)
        def wrapper_alert_error(*args, **kwargs):
            try:
                value = func(*args, **kwargs)
            except Exception as e:
                alert.add_msg(f'{e}', type_='error')
                if debug:
                    raise e
            return value
        return wrapper_alert_error
    return decorator_alert_error

def need_ee(func):
    """
    Decorator to execute check if the object require EE binding.
    Trigger an exception if the connection is not possible. 
    
    Params:
        func (obj): the object on which the decorator is applied
    """
    @wraps(func)
    def wrapper_ee(*args, **kwargs):
        
        # try to connect to ee 
        try: 
            init_ee()
        except Exception as e:
            raise Exception ('This function needs an Earth Engine authentication')
            
        return func(*args, **kwargs)
        
    return wrapper_ee

def loading_button(alert=None, button=None, debug=False):
    """
    Decorator to execute try/except sentence and toggle loading button object.
    Designed to work within the Tile object, or any object that have a self.btn and self.alert set.
    
    Params:
        button (sw.Btn, optional): Toggled button
        alert (sw.Alert, optional): the alert to display the error message
        debug (bool, optional): wether or not the exception should stop the execution. default to False
    """
    
    def decorator_loading(func):
        
        @wraps(func)
        def wrapper_loading(self, *args, **kwargs):
            
            # set btn and alert
            # Change name of variable to assign it again in this scope
            button_ = self.btn if not button else button
            alert_ = self.alert if not alert else alert

            button_.toggle_loading() # Start loading 
            value = None
            try:
                # Catch warnings in the process function
                with warnings.catch_warnings(record=True) as w:
                    value = func(self, *args, **kwargs)
                
                # Check if there are warnings in the function and append them
                # Use append msg due to several warnings could be triggered
                if w: [
                    alert_.append_msg(warning.message.args[0], type_='warning') 
                    for warning in w
                ]
                  
            except Exception as e:
                alert_.add_msg(f'{e}', 'error')
                if debug:
                    button_.toggle_loading() # Stop loading button if there is an error
                    raise e
                    
            button_.toggle_loading() # Stop loading button
            
            return value
        return wrapper_loading
    return decorator_loading


def normalize_str(msg, folder=True):
    """
    Normalize an str to make it compatible with file naming (no spaces, special chars ...etc)
    
    Params:
        msg (str): the string to sanitise
        folder (optional|bool): if the name will be used for folder naming or for display. if display, <'> and < > characters will be kept 
        
    Return:
        (str): the modified str
    """
    
    regex = '[^a-zA-Z\d\-\_]' if folder else '[^a-zA-Z\d\-\_\ \']'
    
    return re.sub(regex, '_', unidecode(msg))
    


import json
from pathlib import Path
from traitlets import Any
from urllib.request import urlretrieve

import pandas as pd
import geopandas as gpd
from ipyleaflet import GeoJSON
import geemap
import ee

from sepal_ui.frontend.styles import AOI_STYLE
from sepal_ui.scripts import utils as su
from sepal_ui.scripts import gee
from sepal_ui.message import ms
from sepal_ui.model import Model


class AoiModel(Model):
    """
    an Model object dedicated to the sorage and the manipulation of aoi. 
    It is meant to be used with the AoiView object (embeded in the AoiTile). 
    By using this you will be able to provide your application with aoi as an ee_object or a gdf, depending if you activated the ee binding or not.
    The class also provide insight on your aoi geometry.
    
    Args: 
        gee (bool, optional): wether or not the aoi selector should be using the EarthEngine binding
        vector (str|pathlib.Path, optional): the path to the default vector object
        admin (int, optional): the administrative code of the default selection. Need to be GADM if ee==False and GAUL 2015 if ee==True.
        asset (str, optional): the default asset. Can only work if ee==True
        folder(str, optional): the init GEE asset folder where the asset selector should start looking (debugging purpose)
        
    Attributes:
        method (traitlet, str): the method to use in the aoi selection. Need to be one of the AoiView.
        
        default_vector (str|pathlib.Path): the default vector file 
        default asset (str): the default asset path
        default_admin (int): the default admin code in GADM if ee==False else GAUL 2015
        
        point_json (traitlet, dict): information that will be use to transform the csv into a gdf (pathname, lat, long, id)
        vector_json (trailtet, dict): information that will be use to transform the vector file into a gdf
        geo_json (traitlet, GeoJson): the drawn geojson featureCollection
        admin (traitlet, int): the administrative code. Need to be GADM if ee==False and GAUL 2015 if ee==True. 
        asset_name (traitlet, str): the asset name (only for GEE model)
        name (traitlet, str): the name of the asset (manually set only in drawn shaped, automatic for all the others)
        
        feature_collection (ee.FeatureCollection): the featurecollection of the aoi (only in ee model)
        gdf (geopandas.GeoDataFrame): the geopandas representation of the aoi
        ipygeojson (GeoJson layer): the ipyleaflet representation of the selected aoi
    """
    
    # const params
    FILE = [Path(__file__).parents[1]/'scripts'/'gadm_database.csv', Path(__file__).parents[1]/'scripts'/'gaul_database.csv']
    CODE = ['GID_{}', 'ADM{}_CODE']
    NAME = ['NAME_{}', 'ADM{}_NAME']
    ISO = ['GID_0', 'ISO 3166-1 alpha-3']
    GADM_BASE_URL = "https://biogeo.ucdavis.edu/data/gadm3.6/gpkg/gadm36_{}_gpkg.zip" # the base url to download gadm maps
    GADM_ZIP_DIR = Path('~', 'tmp', 'GADM_zip').expanduser() # the zip dir where we download the zips
    GADM_ZIP_DIR.mkdir(parents=True, exist_ok=True)
    GAUL_ASSET = "FAO/GAUL/2015/level{}"
    ASSET_SUFFIX = 'aoi_' # the suffix to identify the asset in GEE
    
    # const methods 
    CUSTOM = ms.aoi_sel.custom
    ADMIN = ms.aoi_sel.administrative
    METHODS = {
        'ADMIN0': {'name': ms.aoi_sel.adm[0], 'type': ADMIN},
        'ADMIN1': {'name': ms.aoi_sel.adm[1], 'type': ADMIN},
        'ADMIN2': {'name': ms.aoi_sel.adm[2], 'type': ADMIN},
        'SHAPE': {'name': ms.aoi_sel.vector, 'type': CUSTOM},
        'DRAW': {'name': ms.aoi_sel.draw, 'type': CUSTOM},
        'POINTS': {'name': ms.aoi_sel.points, 'type': CUSTOM},
        'ASSET': {'name': ms.aoi_sel.asset, 'type': CUSTOM}
    }
    
    # widget related traitlets
    method = Any(None).tag(sync=True)
    
    point_json = Any(None).tag(sync=True) # information that will be use to transform the csv into a gdf
    vector_json = Any(None).tag(sync=True) # information that will be use to transform the vector file into a gdf
    geo_json = Any(None).tag(sync=True) # the drawn geojson featureCollection
    admin = Any(None).tag(sync=True)
    asset_name = Any(None).tag(sync=True) # the asset name (only for GEE model)
    name = Any(None).tag(sync=True) # the name of the file (use only in drawn shaped)

    def __init__(self, alert, gee=True, vector=None, admin=None, asset=None, folder=None):

        super().__init__()
        
        # the ee retated informations
        self.ee = gee 
        if gee:
            su.init_ee()
            self.folder = folder if folder else ee.data.getAssetRoots()[0]['id']
        
        # outputs of the selection
        self.gdf = None
        self.feature_collection = None
        self.ipygeojson = None
        
        # the Alert used to display information
        self.alert = alert
        
        # set default values
        self.set_default(vector, admin, asset)
        
        
    def set_default(self, vector=None, admin=None, asset=None):
        """
        Set the default value of the object and create a gdf/feature_collection out of it
        
        Params:
            vector (str|pathlib.path, optional): the default vector file that 
                will be used to produce the gdf. need to be readable by fiona and/or GDAL/OGR
            admin (str, optional): the default administrative area in GADM or GAUL norm
            asset (str, optional): the default asset name, need to point to a readable FeatureCollection
            
        Return:
            self
        """
        
        # save the default values
        self.default_vector = vector
        self.default_asset = self.asset_name = asset
        self.default_admin = self.admin = admin
        
        # cast the vector to json
        self.vector_json = {
            'pathname': str(vector), 
            'column': 'ALL', 
            'value': None
        } if vector else None
        
        # set the default gdf in possible 
        if self.vector_json != None:
            self.set_object('SHAPE')
        elif self.admin != None:
            self.set_object('ADMIN0') # any level will work
        elif self.asset_name != None:
            self.set_object('ASSET')
            
        return self
    
    def set_object(self, method=None):
        """
        set the object (gdf/featurecollection) based on the model inputs. The method can be manually overwrite
        
        Args:
            method (str, optional): a model loading method
        
        Return:
            self
        """
        
        # overwrite self.method
        self.method = method or self.method
        
        if self.method in ['ADMIN0', 'ADMIN1', 'ADMIN2']:
            self._from_admin(self.admin)
        elif self.method == 'POINTS':
            self._from_points(self.point_json)
        elif self.method == 'SHAPE':
            self._from_vector(self.vector_json)
        elif self.method == 'DRAW':
            self._from_geo_json(self.geo_json)
        elif self.method == 'ASSET':
            self._from_asset(self.asset_name)
        else:
            raise Exception(ms.aoi_sel.no_inputs)
            
        self.alert.add_msg(ms.aoi_sel.complete, "success")
        
        return self
    
    def _from_asset(self, asset_name):
        """set the ee.FeatureCollection output from an existing asset"""
        
        # check that I have access to the asset 
        ee_col = ee.FeatureCollection(asset_name)
        ee_col.geometry().bounds().coordinates().get(0).getInfo() # it will raise and error if we cannot access the asset
        
        # set the feature collection 
        self.feature_collection = ee_col 
        
        # create a gdf form te feature_collection 
        self.gdf = geemap.ee_to_geopandas(self.feature_collection)
        
        # set the name 
        self.name = Path(asset_name).stem.replace(self.ASSET_SUFFIX, '')
        
        return self

    def _from_points(self, point_json):
        """set the object output from a csv json"""
        
        if not all(point_json.values()):
            raise Exception('All fields are required, please fill them.')
            
        # cast the pathname to pathlib Path
        point_file = Path(point_json['pathname'])
    
        # check that the columns are well set 
        values = [v for v in point_json.values()]
        if not len(values) == len(set(values)):
            raise Exception(ms.aoi_sel.duplicate_key)
    
        # create the gdf
        df = pd.read_csv(point_file, sep=None, engine='python')
        self.gdf = gpd.GeoDataFrame(
            df, 
            crs='EPSG:4326', 
            geometry = gpd.points_from_xy(
                df[point_json['lng_column']], 
                df[point_json['lat_column']])
        )
        
        # set the name
        self.name = point_file.stem
        
        if self.ee:
            # transform the gdf to ee.FeatureCollection 
            self.feature_collection = geemap.geojson_to_ee(self.gdf.__geo_interface__)

            # export as a GEE asset
            self.export_to_asset()
            
        return self
    
    def _from_vector(self, vector_json):
        """set the object output from a vector json"""
        
        if not (vector_json['pathname']):
            raise Exception('Please select a file.')
        
        if vector_json['column'] != 'ALL':
            if vector_json['value'] is None:
                raise Exception('Please select a value.')
            
        # cast the pathname to pathlib Path
        vector_file = Path(vector_json['pathname'])
        
        # create the gdf
        self.gdf = gpd.read_file(vector_file).to_crs("EPSG:4326")
        
        # set the name using the file stem
        self.name = vector_file.stem
        
        # filter it if necessary
        if vector_json['value']:
            self.gdf = self.gdf[self.gdf[vector_json['column']] == vector_json['value']]
            self.name = f"{self.name}_{vector_json['column']}_{vector_json['value']}"
        
        if self.ee:
            # transform the gdf to ee.FeatureCollection 
            self.feature_collection = geemap.geojson_to_ee(self.gdf.__geo_interface__)

            # export as a GEE asset
            self.export_to_asset()
        
        return self
    
    def _from_geo_json(self, geo_json):
        """set the gdf output from a geo_json"""
        
        if not geo_json:
            raise Exception('Please draw a shape in the map')
        
        # remove the style property from geojson as it's not recognize by geopandas and gee
        for feat in geo_json['features']:
            del feat['properties']['style']
        
        # create the gdf
        self.gdf = gpd.GeoDataFrame.from_features(geo_json)
        
        # normalize the name
        self.name =su.normalize_str(self.name)
        
        if self.ee:
            # transform the gdf to ee.FeatureCollection 
            self.feature_collection = geemap.geojson_to_ee(self.gdf.__geo_interface__)

            # export as a GEE asset
            self.export_to_asset()
        else:
            # save the geojson in downloads 
            path = Path('~', 'downloads', 'aoi').expanduser()
            path.mkdir(exist_ok=True, parents=True) # if nothing have been run the downloads folder doesn't exist
            self.gdf.to_file(path/f'{self.name}.geojson', driver='GeoJSON')
        
        return self
            
    def _from_admin(self, admin):
        """Set the object according to given an administrative number in the GADM norm. The object will be projected in EPSG:4326"""
        
        if not admin:
            raise Exception('Select an administrative layer')
        
        # get the admin level corresponding to the given admin code
        df = pd.read_csv(self.FILE[self.ee])
        
        # extract the first element that include this administrative code and set the level accordingly 
        is_in = df.filter([self.CODE[self.ee].format(i) for i in range(3)]).isin([admin])
        
        if not is_in.any().any():
            raise Exception("The code is not in the database")
        else:
            index = 3 if self.ee else -1
            level = is_in[~((~is_in).all(axis=1))].idxmax(1).iloc[0][index] # the character that contains the index
            
        if self.ee:
            
            # get the feature_collection
            self.feature_collection = ee.FeatureCollection(self.GAUL_ASSET.format(level)) \
                .filter(ee.Filter.eq(f'ADM{level}_CODE', admin))
            
            # transform it into gdf
            self.gdf = geemap.ee_to_geopandas(self.feature_collection)
            
        else:     
            # save the country iso_code 
            iso_3 = admin[:3]
        
            # download the geopackage in tmp 
            zip_file = self.GADM_ZIP_DIR/f'{iso_3}.zip'
        
            if not zip_file.is_file():
            
                # get the zip from GADM server only the ISO_3 code need to be used
                urlretrieve(self.GADM_BASE_URL.format(iso_3), zip_file)
            
            # read the geopackage 
            layer_name = f"gadm36_{iso_3}_{level}"
            level_gdf = gpd.read_file(f'{zip_file}!gadm36_{iso_3}.gpkg', layer=layer_name)
        
            # get the exact admin from this layer 
            self.gdf = level_gdf[level_gdf[self.CODE[self.ee].format(level)] == admin]
        
        # set the name using the layer 
        r = df[df[self.CODE[self.ee].format(level)] == admin].iloc[0]
        names = [su.normalize_str(r[self.NAME[self.ee].format(i)]) if i else r[self.ISO[self.ee]] for i in range(int(level)+1)]
        self.name = '_'.join(names)
        
        return self
    
    def clear_attributes(self):
        """
        Return all attributes to their default state.
        Set the default setting as current object.

        Return: 
            self
        """

        # keep the default 
        admin = self.default_admin
        vector = self.default_vector 
        asset = self.default_asset

        # delete all the traits
        [setattr(self, attr, None) for attr in self.trait_names()]
        
        # reset the outputs
        self.gdf = None
        self.feature_collection = None
        self.ipygeojson = None
        self.selected_feature = None

        # reset the default 
        self.set_default(vector, admin, asset)

        return self

    def get_columns(self):
        """ 
        Retrieve the columns or variables from self excluding geometries and gee index.

        Return: 
            ([str]): sorted list of column names
        """
        
        if type(self.gdf) == type(None):
            raise Exception("You must set the gdf before interacting with it")
        
        if self.ee: 
            aoi_ee = ee.Feature(self.feature_collection.first())
            columns = aoi_ee.propertyNames().getInfo()
            list_ = [col for col in columns if col not in ['system:index', 'Shape_Area']]
        else:
            list_ = list(set(['geometry'])^set(self.gdf.columns.to_list()))
            
        return sorted(list_)
        
    def get_fields(self, column):
        """" 
        Retrieve the fields from a column
        
        Args:
            column (str): A column name to query over the asset
        
        Return: 
            ([str]): sorted list of fields value

        """
        
        if type(self.gdf) == type(None):
            raise Exception("You must set the gdf before interacting with it")
        
        if self.ee: 
            fields = self.feature_collection.distinct(column).aggregate_array(column)
            list_ = fields.getInfo()
        else:
            list_ = self.gdf[column].to_list()
            
        return sorted(list_)
    
    def get_selected(self, column, field):
        """ 
        Select an ee object based on selected column and field.

        Return:
            (ee.Feature|GoeSeries): the Feature associated with the query
        """
        
        if type(self.gdf) == type(None):
            raise Exception("You must set the gdf before interacting with it") 
        
        if self.ee:
            selected_feature = self.feature_collection.filterMetadata(column, 'equals', field)
        else:
            selected_feature = self.gdf[self.gdf[column] == field]
            
        return selected_feature
    
    def total_bounds(self):
        """
        Reproduce the behaviour of the total_bounds method from geopandas
        
        Return: 
            (tuple): minxx, miny, maxx, maxy
        """
        
        if self.ee:
            ee_bounds = self.feature_collection.geometry().bounds().coordinates()
            coords = ee_bounds.get(0).getInfo()
            ll, ur = coords[0], coords[2]
            bounds = ll[0], ll[1], ur[0], ur[1]
        else: 
            bounds = self.gdf.total_bounds
        
        return bounds
    
    def export_to_asset(self):
        """
        Export the feature_collection as an asset (only for ee model)
        
        Return: 
            self
        """
        
        asset_name = self.ASSET_SUFFIX + self.name
        asset_id = str(Path(self.folder, asset_name))
        
        # set the asset name 
        self.asset_name = asset_id
        
        # check if the table already exist
        if asset_id in [a['name'] for a in gee.get_assets(self.folder)]:
            return self
        
        # check if the task is running 
        if gee.is_running(asset_name):
            return self
        
        # run the task
        task_config = {
            'collection': self.feature_collection,
            'description': asset_name,
            'assetId': asset_id
        }
        
        task = ee.batch.Export.table.toAsset(**task_config)
        task.start()
        
        return self
    
    def get_ipygeojson(self):
        """ 
        Converts current geopandas object into ipyleaflet GeoJSON
        
        Return: 
            (GeoJSON): the geojson layer of the aoi gdf
        """
        
        if type(self.gdf) == type(None):
            raise Exception("You must set the gdf before converting it into GeoJSON")
    
        data = json.loads(self.gdf.to_json())
        self.ipygeojson = GeoJSON(data=data, style=AOI_STYLE, name='aoi')
        
        return self.ipygeojson
        


from traitlets import HasTraits
import json

from ipywidgets import dlink

class Model(HasTraits):
    """
    Model object to store information of your computation (input, output.. etc)
    The Model structure is based on traitlets and embed function for export or import in json format. 
    The model traintlets can be bind to any widget field. 
    
    Args:
        traits ([str], optional): a list of traits names (not functional). default to none
        
    """

    def __init__(self, traits=[], *args, **kwargs):
        
        # didn't manage to make it work and it seems dodgy has it creates a simlink to new class at each call
        #if len(traits):
        #    [self.add_traits(Any(val).tag(sync=True), label) for label, val in traits.items()]
                
        
        super().__init__(*args, **kwargs)
        
    def __repr__(self):
        """Method to represent the Model objects as a string"""
        
        args = ", ".join([f'{k}={v}' for k, v in self.export_data().items()]).strip()
        
        return f'{self.__class__.__name__}({args})'
        
        
    def export_data(self):
        """
        Export a json description of all the object traitlets 
        Note that the members will be ignored
        
        Return:
            (dict): the serialized traitlets
        """
        
        return self.__dict__['_trait_values']
    
    def import_data(self, data):
        """
        Import the model (i.e. the traitlets) from a json file
        
        Params:
            json (dict): the json ditionary of the model
            
        Return:
            self
        """
        
        if type(data) == str:
            data = json.loads(data)
        
        for k, v in data.items():
            getattr(self, k) # to raise an error if the json is malformed 
            setattr(self, k, v)
            
    def bind(self, widget, trait):
        """
        Binding a widget input 'v_model' trait to a trait of the model. 
        The binding will be unidirectionnal for the sake of some custom widget that does not support it.
        This wrapper avoid to import the ipywidgets lib everywhere and reduce the number of parameters
        Some existence check are also performed and will throw an error if the trait doesn't exist
        
        Params:
            widget (v.widget): any input widget with a v_model trait
            trait (str): the name of a trait of the current model
            
        Return:
            self
        """
        
        # check trait existence 
        getattr(self, trait)
        
        # bind them 
        dlink((widget, 'v_model'),(self, trait))
        
        # maybe I would add the possiblity to add an alert to display stuff to the user with the same options as in alert.bind
        
        return self
        
        
        